/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.IO;
using System.Reflection;

namespace Squarius2 {

  public class Point2D {
    public int x;
    public int y;
    public Point2D() {
    }
    public Point2D(int x, int y) {
      this.x = x;
      this.y = y;
    }
  }

  public enum BulletType { Player, Enemy };

  public class Bullet {
    public double x;
    public double y;
    public double vx;
    public double vy;
    public SquariusForm squariusForm;
    public BulletType bulletType;

    public Bullet(SquariusForm squariusForm, BulletType bulletType,
        double x, double y) {
      this.squariusForm = squariusForm;
      this.bulletType = bulletType;
      this.x = x;
      this.y = y;
    }

    public Bullet(SquariusForm squariusForm, BulletType bulletType,
        double x, double y, double vx, double vy) 
            : this(squariusForm, bulletType, x, y) {
      this.vx = vx;
      this.vy = vy;
    }

    public Bullet(SquariusForm squariusForm, Ship enemy) 
        : this(squariusForm, BulletType.Enemy, 
              enemy.x + (enemy.width >> 1),
              enemy.y + (enemy.height >> 1)) {
      Ship playerShip = squariusForm.playerShip;
      double targetX = playerShip.x + (playerShip.width >> 1);
      double targetY = playerShip.y + (playerShip.height >> 1);
      vx = (targetX - x);
      vy = (targetY - y);
      double scale = 5 / Math.Sqrt(vx * vx + vy * vy);
      vx *= scale;
      vy *= scale;
      vx += squariusForm.scrollSpeed;
    }

    public int X {
      get {
        return (int)(x - 4);
      }
    }
    public int Y {
      get {
        return (int)(y - 4);
      }
    }
    public Brush Brush {
      get {
        return (bulletType == BulletType.Player) 
            ? Brushes.LightGreen : Brushes.Yellow;
      }
    }

    public bool Move() {
      x += vx;
      y += vy;
      if ((y - 5) < 0 || (y + 5) >= 480
          || (x - 5) < squariusForm.viewportX
          || (x + 5) >= squariusForm.viewportX + 640) {
        return true;
      }
      if (squariusForm.IsWall((int)x, (int)y)
          && squariusForm.fireworks == 0) {
        squariusForm.explosions.Add(new Explosion(squariusForm, (int)x, (int)y, 20));
        return true;
      }
      if (bulletType == BulletType.Player) {
        List<Ship> ships = squariusForm.ships;
        for (int i = ships.Count - 1; i >= 0; i--) {
          Ship ship = ships[i];
          if (ship.ContainsPoint((int)x, (int)y)) {
            if (ship.Shot()) {
              squariusForm.explosions.Add(new Explosion(squariusForm, ship));
              squariusForm.squaresDestroyed++;
              ships.RemoveAt(i);
              if (ship.yieldsPowerUp) {
                squariusForm.powerUps.Add(new PowerUp(squariusForm, ship));
              }
            } else {
              squariusForm.explosions.Add(
                  new Explosion(squariusForm, (int)x, (int)y, 20));
            }
            return true;
          }
        }
      } else if (squariusForm.playerAlive) {
        Ship playerShip = squariusForm.playerShip;
        if (squariusForm.weaponLevel > 5) {          
          int dx = playerShip.x + 16 - (int)x;
          int dy = playerShip.y + 16 - (int)y;
          if (dx * dx + dy * dy <= 512) {
            squariusForm.weaponLevel--;
            squariusForm.explosions.Add(
                  new Explosion(squariusForm, (int)x, (int)y, 20));
            return true;
          }
        } else if (playerShip.ContainsPoint((int)x, (int)y)) {
          squariusForm.PlayerKilled();
          return true;
        }
      }
      return false;
    }
  }

  public class Explosion {
    public int x;
    public int y;
    public Brush brush = Brushes.White;
    public int radius;
    public int maxRadius;
    public SquariusForm squariusForm;

    public Explosion(SquariusForm squariusForm, int x, int y)
      : this(squariusForm, x, y, 50) {
    }

    public Explosion(SquariusForm squariusForm, int x, int y, int maxRadius) {
      this.squariusForm = squariusForm;
      this.x = x;
      this.y = y;
      this.maxRadius = maxRadius;
    }

    public Explosion(SquariusForm squariusForm, Ship ship)
      : this(
        squariusForm,
        ship.x + (ship.width >> 1),
        ship.y + (ship.height >> 1)) {
    }

    public bool Move() {
      radius += 10;
      brush = (brush == Brushes.White) ? Brushes.Yellow : Brushes.White;
      return radius > maxRadius;
    }
  }

  public class Ship {
    public int x = 64;
    public int y = 224;
    public int width = 32;
    public int height = 32;
    public Brush brush = Brushes.Red;
    public SquariusForm squariusForm;
    public bool yieldsPowerUp;

    public Ship(SquariusForm squariusForm) {
      this.squariusForm = squariusForm;
    }

    public bool ContainsPoint(int X, int Y) {
      return X >= x && Y >= y && X < x + width && Y < y + height;
    }

    public bool ContainsPoints(Ship ship) {
      return ContainsPoint(ship.x, ship.y)
          || ContainsPoint(ship.x + ship.width, ship.y)
          || ContainsPoint(ship.x, ship.y + ship.height)
          || ContainsPoint(ship.x + ship.width, ship.y + ship.height);
    }

    public bool Intersects(Ship ship) {
      return ContainsPoints(ship) || ship.ContainsPoints(this);
    }

    public virtual bool Move() {
      return false;
    }

    public virtual bool Shot() {
      return false;
    }
  }

  public class PowerUp : Ship {

    public int brushCount;

    public PowerUp(SquariusForm squariusForm, int x, int y)
        : base(squariusForm) {
      this.x = x;
      this.y = y;
      width = height = 20;
      brush = Brushes.Purple;
    }

    public PowerUp(SquariusForm squariusForm, Ship enemy) 
        : this(squariusForm, 
            enemy.x + (enemy.width >> 1) - 10,
            enemy.y + (enemy.height >> 1) - 10) {
    }

    public override bool Move() {
      if (brushCount == 0) {
        brush = (brush == Brushes.Purple) ? Brushes.White : Brushes.Purple;
        brushCount = 2;
      } else {
        brushCount--;
      }
      if (x + 20 < squariusForm.viewportX) {
        return true;
      }
      if (squariusForm.playerShip.Intersects(this)) {
        if (squariusForm.weaponLevel < 8) {
          squariusForm.weaponLevel++;
        }
        squariusForm.weaponsCollectedOnStage++;
        return true;
      }
      return false;
    }
  }
}