/*
 * SQUARIUS^2
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of SQUARIUS^2.
 *
 * SQUARIUS^2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SQUARIUS^2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.IO;
using System.Reflection;

namespace Squarius2 {
  public partial class SquariusForm : Form {

    [StructLayout(LayoutKind.Sequential)]
    public struct PeekMsg {
      public IntPtr hWnd;
      public Message msg;
      public IntPtr wParam;
      public IntPtr lParam;
      public uint time;
      public System.Drawing.Point p;
    }

    [System.Security.SuppressUnmanagedCodeSecurity]
    [DllImport("User32.dll", CharSet = CharSet.Auto)]
    public static extern bool PeekMessage(out PeekMsg msg, IntPtr hWnd,
             uint messageFilterMin, uint messageFilterMax, uint flags);

    public enum GameMode { Introduction, Playing, Ending }

    public const int FRAMES_PER_SECOND = 40;
    public const int STAGES = 8;
    public bool paused = false;
    public bool upKeyPressed;
    public bool downKeyPressed;
    public bool leftKeyPressed;
    public bool rightKeyPressed;
    public bool enterKeyPressed;
    public bool shootKeyPressed;
    public Random random = new Random();
    public Bitmap bitmap;
    public long FREQUENCY = Stopwatch.Frequency;
    public long TICKS_PER_FRAME = Stopwatch.Frequency / FRAMES_PER_SECOND;
    public Stopwatch stopwatch = new Stopwatch();
    public long nextFrameStart;
    public Graphics g;
    public Graphics graphics;
    public GameMode gameMode = GameMode.Introduction;
    public char[,][] stages = new char[STAGES, 15][];
    public Brush wallBrush = Brushes.SandyBrown;
    public int stage;
    public int viewportX;
    public int lastViewPortTrigger;
    public int scrollSpeed;
    public int[][,] bulletMaps = new int[2][,];
    public bool deactivationPause;
    public int attempts;
    public int mouseCursorTimeout = 20;
    public bool mouseCursorVisible = true;
    public bool mouseEntered;
    public bool fullScreenMode;

    public SquariusForm() {
      InitializeComponent();
      ClientSize = new Size(640, 480);
      StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
      SetStyle(ControlStyles.UserPaint, true);
      SetStyle(ControlStyles.AllPaintingInWmPaint, true);
    }

    private void SquariusForm_Load(object sender, EventArgs e) {
      playerShip = new Ship(this);
      LoadStages();
      CreateBulletMaps();                      
      bitmap = new Bitmap(640, 480, graphics = CreateGraphics());
      g = Graphics.FromImage(bitmap);
      StartIntroduction();
      stopwatch.Start();
      nextFrameStart = stopwatch.ElapsedTicks + TICKS_PER_FRAME;
      Application.Idle += OnApplicationIdle;
    }

    public void StartIntroduction() {
      gameMode = GameMode.Introduction;
      InitializeIntroduction();
    }

    public void RestartStage() {
      InitStage();
      RestorePlayer();
    }

    public void StartPlaying() {
      gameMode = GameMode.Playing;      
      stage = 0;
      attempts = 1;
      squaresDestroyed = 0;
      InitStage(); 
      RestorePlayer();       
    }

    public void StartEnding() {
      gameMode = GameMode.Ending;
      InitializeEnding();
    }

    private void OnApplicationIdle(object sender, EventArgs e) {
      PeekMsg msg;
      while (!PeekMessage(out msg, IntPtr.Zero, 0, 0, 0)) {
        do {
          if (!paused) {
            UpdateModel();
          }
          nextFrameStart += TICKS_PER_FRAME;
        } while (nextFrameStart < stopwatch.ElapsedTicks);
        RenderFrame();
        graphics.DrawImageUnscaled(bitmap, 0, 0);
        long remainingTicks = nextFrameStart - stopwatch.ElapsedTicks;
        if (remainingTicks > 0) {
          Thread.Sleep((int)(1000 * remainingTicks / FREQUENCY));
        }
      }
    }

    private void UpdateModel() {
      if (mouseCursorTimeout > 0) {
        mouseCursorTimeout--;
        if (mouseCursorTimeout == 0) {
          HideMouseCursor();
        }
      }
      switch (gameMode) {
        case GameMode.Introduction:
          UpdateModel_Introduction();
          break;
        case GameMode.Playing:
          UpdateModel_Playing();
          break;
        case GameMode.Ending:
          UpdateModel_Ending();
          break;
      }
    }

    private void RenderFrame() {
      switch (gameMode) {
        case GameMode.Introduction:
          RenderFrame_Introduction();
          break;
        case GameMode.Playing:
          RenderFrame_Playing();
          break;
        case GameMode.Ending:
          RenderFrame_Ending();
          break;
      }
    }

    private void SquariusForm_KeyDown(object sender, KeyEventArgs e) {
      if (!paused) {
        KeyChanged(e, true);
      }
    }

    private void SquariusForm_KeyUp(object sender, KeyEventArgs e) {
      if (!paused) {
        KeyChanged(e, false);
        if (e.KeyCode == Keys.Back) {
          requestRestartStage = true;
        }
      }
      if (e.KeyCode == Keys.P) {
        Paused = !paused;
      } else if (e.KeyCode == Keys.F11) {
        ToggleFullScreenMode();
      } else if (e.KeyCode == Keys.Escape && fullScreenMode) {
        ToggleFullScreenMode();
        Paused = true;
      }
    }

    private bool Paused {
      set {
        if (paused == value) {
          return;
        }
        paused = value;
        if (paused) {
          Text += " [PAUSED]";
          upKeyPressed = false;
          downKeyPressed = false;
          leftKeyPressed = false;
          rightKeyPressed = false;
          enterKeyPressed = false;
          shootKeyPressed = false;
        } else {
          Text = Text.Substring(0, Text.Length - " [PAUSED]".Length);
          deactivationPause = false;
        }
      }
    }

    private void KeyChanged(KeyEventArgs e, bool pressed) {
      switch (e.KeyCode) {
        case Keys.Up:
          upKeyPressed = pressed;
          break;
        case Keys.Down:
          downKeyPressed = pressed;
          break;
        case Keys.Left:
          leftKeyPressed = pressed;
          break;
        case Keys.Right:
          rightKeyPressed = pressed;
          break;
        case Keys.D:          
          shootKeyPressed = pressed;
          break;
        case Keys.Enter:
          enterKeyPressed = pressed;
          break;
      }
    }

    private void SquariusForm_FormClosing(object sender, FormClosingEventArgs e) {
    }

    private void LoadStages() {
      StringBuilder[] sb = new StringBuilder[15];
      for (int j = 0; j < sb.Length; j++) {
        sb[j] = new StringBuilder();
      }

      Assembly assembly = Assembly.GetExecutingAssembly();
      for (int i = 0; i < STAGES; i++) {
        for (int j = 0; j < sb.Length; j++) {
          sb[j].Length = 0;
        }
        int row = 0;
        StreamReader reader = new StreamReader(assembly.GetManifestResourceStream(
            "Squarius2.stages." + i + ".txt"));
        string input = null;
        while ((input = reader.ReadLine()) != null) {
          input = input.Trim();
          if (input.Length == 0) {
            continue;
          }
/*#if DEBUG
          if (input.Length != 20) {
            Console.WriteLine("Bad line ({0}): {1}", i, input);
            Application.Exit();
          }
#endif*/
          sb[row].Append(input);
          if (++row == sb.Length) {
            row = 0;
          }
        }
        reader.Close();

        for (int j = 0; j < sb.Length; j++) {
          stages[i, j] = sb[j].ToString().ToCharArray();
        }
      }
    }

    private void CreateBulletMaps() {
      CreateBulletMap(0, 20, 8);
      CreateBulletMap(1, 33, 16);
    }

    private void CreateBulletMap(int index, double radius, int space) {
      List<Point2D> points = new List<Point2D>();
      for (int y = -space; y <= 32 + space; y += 8) {
        double X = 16 - space;
        while (InsideCircle(8, 16, radius, X, y)) {
          X += 0.1;
        }
        do {
          points.Add(new Point2D((int)X, y));
          X += 8;
        } while (InsideCircle(16 + space, 16, radius, X, y));
      }
      bulletMaps[index] = new int[points.Count, 2];
      for (int i = 0; i < points.Count; i++) {
        Point2D point = points[i];
        bulletMaps[index][i, 0] = point.x;
        bulletMaps[index][i, 1] = point.y;
      }
    }

    public bool InsideCircle(double cx, double cy, double r, double x, double y) {
      double dx = x - cx;
      double dy = y - cy;
      return dx * dx + dy * dy <= r * r; 
    }

    private void SquariusForm_Activated(object sender, EventArgs e) {
      if (deactivationPause) {
        Paused = false;        
      }
    }

    private void SquariusForm_Deactivate(object sender, EventArgs e) {
      if (!paused) {
        Paused = true;
        deactivationPause = true;
        if (fullScreenMode) {
          fullScreenMode = false;
          ResolutionManager.ExitFullScreenMode(this);
        }
      }
    }

    private void SquariusForm_MouseEnter(object sender, EventArgs e) {
      mouseEntered = true;
      ShowMouseCursor();
    }

    private void SquariusForm_MouseLeave(object sender, EventArgs e) {
      mouseEntered = false;
      ShowMouseCursor();
    }

    private void SquariusForm_MouseMove(object sender, MouseEventArgs e) {
      ShowMouseCursor();      
    }

    private void HideMouseCursor() {
      if (mouseCursorVisible && mouseEntered) {
        mouseCursorVisible = false;
        Cursor.Hide();
      }
    }

    private void ShowMouseCursor() {
      mouseCursorTimeout = 20;
      if (!mouseCursorVisible) {
        mouseCursorVisible = true;
        Cursor.Show();
      }
    }

    private void ToggleFullScreenMode() {
      fullScreenMode = !fullScreenMode;
      if (fullScreenMode) {
        ResolutionManager.EnterFullScreenMode(this, 640, 480);
      } else {
        ResolutionManager.ExitFullScreenMode(this);
      }
      nextFrameStart = stopwatch.ElapsedTicks + TICKS_PER_FRAME;
    }
  }
}